"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var assignment_1 = require("./assignment");
var ts = require("typescript");
var template_1 = require("../template");
var types_1 = require("../types");
var variable_1 = require("./variable");
var regexfunc_1 = require("./regexfunc");
var js_var_operator_map = (_a = {},
    _a[ts.SyntaxKind.AsteriskToken] = "JS_VAR_ASTERISK",
    _a[ts.SyntaxKind.SlashToken] = "JS_VAR_SLASH",
    _a[ts.SyntaxKind.PercentToken] = "JS_VAR_PERCENT",
    _a[ts.SyntaxKind.PlusToken] = "JS_VAR_PLUS",
    _a[ts.SyntaxKind.MinusToken] = "JS_VAR_MINUS",
    _a);
var CBinaryExpression = /** @class */ (function () {
    function CBinaryExpression(scope, node) {
        this.replacedWithCall = false;
        this.replacedWithVar = false;
        this.replacedWithVarAssignment = false;
        this.replacementVarName = null;
        this.gcVarName = null;
        this.strPlusStr = false;
        this.strPlusNumber = false;
        this.numberPlusStr = false;
        this.computeOperation = null;
        this.expression = null;
        if (node.operatorToken.kind == ts.SyntaxKind.EqualsToken) {
            this.expression = assignment_1.AssignmentHelper.create(scope, node.left, node.right, true);
            return;
        }
        if (node.operatorToken.kind == ts.SyntaxKind.CommaToken) {
            var nodeAsStatement = ts.createNode(ts.SyntaxKind.ExpressionStatement);
            nodeAsStatement.expression = node.left;
            nodeAsStatement.parent = node.getSourceFile();
            scope.statements.push(template_1.CodeTemplateFactory.createForNode(scope, nodeAsStatement));
            this.expression = template_1.CodeTemplateFactory.createForNode(scope, node.right);
            return;
        }
        var leftType = scope.root.typeHelper.getCType(node.left);
        this.left = template_1.CodeTemplateFactory.createForNode(scope, node.left);
        var rightType = scope.root.typeHelper.getCType(node.right);
        this.right = template_1.CodeTemplateFactory.createForNode(scope, node.right);
        var operatorKind = node.operatorToken.kind;
        var operatorMap = {};
        var callReplaceMap = {};
        if (leftType == types_1.RegexVarType && operatorKind == ts.SyntaxKind.PlusToken) {
            leftType = types_1.StringVarType;
            this.left = new regexfunc_1.CRegexAsString(this.left);
        }
        if (rightType == types_1.RegexVarType && operatorKind == ts.SyntaxKind.PlusToken) {
            rightType = types_1.StringVarType;
            this.right = new regexfunc_1.CRegexAsString(this.right);
        }
        operatorMap[ts.SyntaxKind.AmpersandAmpersandToken] = '&&';
        operatorMap[ts.SyntaxKind.BarBarToken] = '||';
        if (leftType == types_1.NumberVarType && rightType == types_1.NumberVarType) {
            operatorMap[ts.SyntaxKind.GreaterThanToken] = '>';
            operatorMap[ts.SyntaxKind.GreaterThanEqualsToken] = '>=';
            operatorMap[ts.SyntaxKind.LessThanToken] = '<';
            operatorMap[ts.SyntaxKind.LessThanEqualsToken] = '<=';
            operatorMap[ts.SyntaxKind.ExclamationEqualsEqualsToken] = '!=';
            operatorMap[ts.SyntaxKind.ExclamationEqualsToken] = '!=';
            operatorMap[ts.SyntaxKind.EqualsEqualsEqualsToken] = '==';
            operatorMap[ts.SyntaxKind.EqualsEqualsToken] = '==';
            operatorMap[ts.SyntaxKind.AsteriskToken] = '*';
            operatorMap[ts.SyntaxKind.SlashToken] = '/';
            operatorMap[ts.SyntaxKind.PercentToken] = '%';
            operatorMap[ts.SyntaxKind.PlusToken] = '+';
            operatorMap[ts.SyntaxKind.MinusToken] = '-';
            operatorMap[ts.SyntaxKind.AmpersandToken] = '&';
            operatorMap[ts.SyntaxKind.BarToken] = '|';
            operatorMap[ts.SyntaxKind.CaretToken] = '^';
            operatorMap[ts.SyntaxKind.GreaterThanGreaterThanToken] = '>>';
            operatorMap[ts.SyntaxKind.GreaterThanGreaterThanGreaterThanToken] = '>>';
            operatorMap[ts.SyntaxKind.LessThanLessThanToken] = '<<';
            operatorMap[ts.SyntaxKind.AsteriskEqualsToken] = '*=';
            operatorMap[ts.SyntaxKind.SlashEqualsToken] = '/=';
            operatorMap[ts.SyntaxKind.PercentEqualsToken] = '%=';
            operatorMap[ts.SyntaxKind.PlusEqualsToken] = '+=';
            operatorMap[ts.SyntaxKind.MinusEqualsToken] = '-=';
            operatorMap[ts.SyntaxKind.AmpersandEqualsToken] = '&=';
            operatorMap[ts.SyntaxKind.BarEqualsToken] = '|=';
            operatorMap[ts.SyntaxKind.CaretEqualsToken] = '^=';
            operatorMap[ts.SyntaxKind.GreaterThanGreaterThanEqualsToken] = '>>=';
            operatorMap[ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken] = '>>';
            operatorMap[ts.SyntaxKind.LessThanLessThanEqualsToken] = '<<=';
            if (operatorKind == ts.SyntaxKind.GreaterThanGreaterThanGreaterThanToken
                || operatorKind == ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken) {
                var leftAsString = template_1.CodeTemplateFactory.templateToString(this.left);
                this.left = "((uint16_t)" + leftAsString + ")";
                if (operatorKind == ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken)
                    this.left = leftAsString + " = " + this.left;
                scope.root.headerFlags.uint16_t = true;
            }
        }
        else if (leftType == types_1.StringVarType && rightType == types_1.StringVarType) {
            callReplaceMap[ts.SyntaxKind.ExclamationEqualsEqualsToken] = ['strcmp', ' != 0'];
            callReplaceMap[ts.SyntaxKind.ExclamationEqualsToken] = ['strcmp', ' != 0'];
            callReplaceMap[ts.SyntaxKind.EqualsEqualsEqualsToken] = ['strcmp', ' == 0'];
            callReplaceMap[ts.SyntaxKind.EqualsEqualsToken] = ['strcmp', ' == 0'];
            if (callReplaceMap[operatorKind])
                scope.root.headerFlags.strings = true;
            if (operatorKind == ts.SyntaxKind.PlusToken || operatorKind == ts.SyntaxKind.PlusEqualsToken) {
                var tempVarName = scope.root.memoryManager.getReservedTemporaryVarName(node);
                scope.func.variables.push(new variable_1.CVariable(scope, tempVarName, "char *", { initializer: "NULL" }));
                this.gcVarName = scope.root.memoryManager.getGCVariableForNode(node);
                this.replacedWithVar = true;
                this.replacedWithVarAssignment = operatorKind === ts.SyntaxKind.PlusEqualsToken;
                this.replacementVarName = tempVarName;
                this.strPlusStr = true;
                scope.root.headerFlags.strings = true;
                scope.root.headerFlags.malloc = true;
            }
        }
        else if (leftType == types_1.NumberVarType && rightType == types_1.StringVarType
            || leftType == types_1.StringVarType && rightType == types_1.NumberVarType) {
            callReplaceMap[ts.SyntaxKind.ExclamationEqualsEqualsToken] = ['str_int16_t_cmp', ' != 0'];
            callReplaceMap[ts.SyntaxKind.ExclamationEqualsToken] = ['str_int16_t_cmp', ' != 0'];
            callReplaceMap[ts.SyntaxKind.EqualsEqualsEqualsToken] = ['str_int16_t_cmp', ' == 0'];
            callReplaceMap[ts.SyntaxKind.EqualsEqualsToken] = ['str_int16_t_cmp', ' == 0'];
            if (callReplaceMap[operatorKind]) {
                scope.root.headerFlags.str_int16_t_cmp = true;
                // str_int16_t_cmp expects certain order of arguments (string, number)
                if (leftType == types_1.NumberVarType) {
                    var tmp = this.left;
                    this.left = this.right;
                    this.right = tmp;
                }
            }
            if (operatorKind == ts.SyntaxKind.PlusToken || operatorKind == ts.SyntaxKind.PlusEqualsToken) {
                var tempVarName = scope.root.memoryManager.getReservedTemporaryVarName(node);
                scope.func.variables.push(new variable_1.CVariable(scope, tempVarName, "char *", { initializer: "NULL" }));
                this.gcVarName = scope.root.memoryManager.getGCVariableForNode(node);
                this.replacedWithVar = true;
                this.replacedWithVarAssignment = operatorKind === ts.SyntaxKind.PlusEqualsToken;
                this.replacementVarName = tempVarName;
                if (leftType == types_1.NumberVarType)
                    this.numberPlusStr = true;
                else
                    this.strPlusNumber = true;
                scope.root.headerFlags.strings = true;
                scope.root.headerFlags.malloc = true;
                scope.root.headerFlags.str_int16_t_cat = true;
            }
        }
        else if (leftType == types_1.UniversalVarType || rightType == types_1.UniversalVarType) {
            this.computeOperation = js_var_operator_map[operatorKind];
            this.left = new variable_1.CAsUniversalVar(scope, node.left, this.left, leftType);
            this.right = new variable_1.CAsUniversalVar(scope, node.right, this.right, rightType);
            scope.root.headerFlags.js_var_compute = true;
        }
        this.operator = operatorMap[operatorKind];
        if (callReplaceMap[operatorKind]) {
            this.replacedWithCall = true;
            _a = callReplaceMap[operatorKind], this.call = _a[0], this.callCondition = _a[1];
        }
        this.nodeText = node.flags & ts.NodeFlags.Synthesized ? "(synthesized node)" : node.getText();
        if (this.gcVarName) {
            scope.root.headerFlags.gc_iterator = true;
            scope.root.headerFlags.array = true;
        }
        var _a;
    }
    CBinaryExpression = __decorate([
        template_1.CodeTemplate("\n{#statements}\n    {#if replacedWithVar && strPlusStr}\n        {replacementVarName} = malloc(strlen({left}) + strlen({right}) + 1);\n        assert({replacementVarName} != NULL);\n        strcpy({replacementVarName}, {left});\n        strcat({replacementVarName}, {right});\n    {#elseif replacedWithVar && strPlusNumber}\n        {replacementVarName} = malloc(strlen({left}) + STR_INT16_T_BUFLEN + 1);\n        assert({replacementVarName} != NULL);\n        {replacementVarName}[0] = '\\0';\n        strcat({replacementVarName}, {left});\n        str_int16_t_cat({replacementVarName}, {right});\n    {#elseif replacedWithVar && numberPlusStr}\n        {replacementVarName} = malloc(strlen({right}) + STR_INT16_T_BUFLEN + 1);\n        assert({replacementVarName} != NULL);\n        {replacementVarName}[0] = '\\0';\n        str_int16_t_cat({replacementVarName}, {left});\n        strcat({replacementVarName}, {right});\n    {/if}\n    {#if replacedWithVar && gcVarName}\n        ARRAY_PUSH({gcVarName}, {replacementVarName});\n    {/if}\n\n{/statements}\n{#if expression}\n    {expression}\n{#elseif operator}\n    {left} {operator} {right}\n{#elseif replacedWithCall}\n    {call}({left}, {right}){callCondition}\n{#elseif replacedWithVarAssignment}\n    ({left} = {replacementVarName})\n{#elseif replacedWithVar}\n    {replacementVarName}\n{#elseif computeOperation}\n    js_var_compute({left}, {computeOperation}, {right})\n{#else}\n    /* unsupported expression {nodeText} */\n{/if}", ts.SyntaxKind.BinaryExpression)
    ], CBinaryExpression);
    return CBinaryExpression;
}());
exports.CBinaryExpression = CBinaryExpression;
var CUnaryExpression = /** @class */ (function () {
    function CUnaryExpression(scope, node) {
        this.call = null;
        this.callPostfix = "";
        var operatorMap = {};
        var type = scope.root.typeHelper.getCType(node.operand);
        this.operand = template_1.CodeTemplateFactory.createForNode(scope, node.operand);
        this.isPostfix = node.kind == ts.SyntaxKind.PostfixUnaryExpression;
        this.nodeText = node.getText();
        operatorMap[ts.SyntaxKind.ExclamationToken] = '!';
        if (type == types_1.NumberVarType) {
            operatorMap[ts.SyntaxKind.PlusPlusToken] = '++';
            operatorMap[ts.SyntaxKind.MinusMinusToken] = '--';
            operatorMap[ts.SyntaxKind.MinusToken] = '-';
            operatorMap[ts.SyntaxKind.PlusToken] = '+';
            operatorMap[ts.SyntaxKind.TildeToken] = '~';
        }
        if (type != types_1.NumberVarType && type != types_1.UniversalVarType && node.operator === ts.SyntaxKind.PlusToken) {
            this.call = "str_to_int16_t";
            scope.root.headerFlags.str_to_int16_t = true;
        }
        else if (type == types_1.UniversalVarType && node.operator == ts.SyntaxKind.PlusToken) {
            this.call = "js_var_to_number";
            scope.root.headerFlags.js_var_to_number = true;
        }
        this.operator = operatorMap[node.operator];
    }
    CUnaryExpression = __decorate([
        template_1.CodeTemplate("\n{#if isPostfix && operator}\n    {operand}{operator}\n{#elseif !isPostfix && operator}\n    {operator}{operand}\n{#elseif call}\n    {call}({operand})\n{#else}\n    /* unsupported expression {nodeText} */\n{/if}", [ts.SyntaxKind.PrefixUnaryExpression, ts.SyntaxKind.PostfixUnaryExpression])
    ], CUnaryExpression);
    return CUnaryExpression;
}());
var CTernaryExpression = /** @class */ (function () {
    function CTernaryExpression(scope, node) {
        this.condition = template_1.CodeTemplateFactory.createForNode(scope, node.condition);
        this.whenTrue = template_1.CodeTemplateFactory.createForNode(scope, node.whenTrue);
        this.whenFalse = template_1.CodeTemplateFactory.createForNode(scope, node.whenFalse);
    }
    CTernaryExpression = __decorate([
        template_1.CodeTemplate("{condition} ? {whenTrue} : {whenFalse}", ts.SyntaxKind.ConditionalExpression)
    ], CTernaryExpression);
    return CTernaryExpression;
}());
var CGroupingExpression = /** @class */ (function () {
    function CGroupingExpression(scope, node) {
        this.expression = template_1.CodeTemplateFactory.createForNode(scope, node.expression);
    }
    CGroupingExpression = __decorate([
        template_1.CodeTemplate("({expression})", ts.SyntaxKind.ParenthesizedExpression)
    ], CGroupingExpression);
    return CGroupingExpression;
}());
var _a;
//# sourceMappingURL=expressions.js.map